/* iBurst (TM) compatible driver for 2.6 Linux kernel.
 * based on the original ArrayComm (TM) iBurst (TM) driver.
 * Nicholas Jefferson <nicholas@pythontraining.com.au>
 * 11 May 2005
 *
 * Ported to 2.6 Linux kernel by Nik Trevallyn-Jones.
 * Patches for 2.6.13 Linux kernel by Greg Cockburn and Scott McKenzie.
 * Patch for 2.6.16 Linux kernel by Daniel Burr.
 * Patch for 2.6.17 Linux kernel by Damian Ivereigh
 * Patch for sysfs by Scott McKenzie.
 * Fixes to support new hardware by Shane MacPhillamy.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#include "ib-net.h"

#include <asm/io.h>
#include <linux/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cisreg.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/ds.h>

static int debug = 9;
module_param(debug, int, 0);
#define DEBUG(n, args...) if (debug < (n)) ; else printk(KERN_DEBUG args)

/**
 * interval - poll interval (in milliseconds).
 */
static int interval = 4;
module_param(interval, int, 0);

/**
 * io16 - PCMCIA 16 bit mode.
 */
static int io16 = 0;
module_param(io16, int, 0);

/**
 * enum ib_pcmcia_ifstate - hardware interface state.
 */
enum ib_pcmcia_ifstate
{
	IFSTATE_RESET = 0,
	IFSTATE_WAIT = 1,
	IFSTATE_NORMAL = 2,
};

/**
 * ib_pcmcia_offset - i/o mapped memory offsets.
 */
enum ib_pcmcia_offset
{
	OFFSET_UT_MAGIC1 = 0x00,
	OFFSET_UT_MAGIC2 = 0x01,
	OFFSET_UT_SEQUENCE = 0x02,
	OFFSET_UT_FEEDBACK = 0x03,
	OFFSET_UT_JRX = 0x04,
	OFFSET_UT_ITX = 0x05,
	OFFSET_UT_PACKET = 0x06,
	OFFSET_UT_STATUS = 0x07,
	OFFSET_PC_MAGIC1 = 0x08,
	OFFSET_PC_MAGIC2 = 0x09,
	OFFSET_PC_SEQUENCE = 0x0a,
	OFFSET_PC_FEEDBACK = 0x0b,
	OFFSET_PC_JTX = 0x0c,
	OFFSET_PC_IRX = 0x0d,
	OFFSET_PC_PACKET = 0x0e,
	OFFSET_PC_STATUS = 0x0f,
	OFFSET_UT_ICHUNK = 0x10,
	OFFSET_UT_NCHUNK = 0x11,
	OFFSET_PC_JCHUNK = 0x12,
	OFFSET_PC_NCHUNK = 0x13,
	OFFSET_UT_ETHERNET = 0x1a,
};

/**
 * ib_pcmcia_local_t - PCMCIA device private state.
 * @dev: PCMCIA device state.
 * @node: network node.
 * @mem: i/o mapped memory.
 * @modem: corresponding modem state.
 * @timer: poll hardware.
 * @release: release configuration timer.
 * @ifstate: hardware state.
 * @sequence: hardware sync protocol.
 * @feedback: hardware sync protocol.
 * @rx_echunk: expected chunks (32 bytes) to receive.
 * @rx_ibuf: next index into rx_buf to use.
 * @rx_nbuf: packet length.
 * @tx_echunk: expected chunks (32 bytes) to transmit.
 * @tx_jbuf: next index into tx_buf to use.
 * @irx: most recent index into i/o mapped receive buffer.
 * @jtx: most recent index into i/o mapped transmit buffer.
 * @nreset: number of hardware resets.
 */
struct ib_pcmcia_local_t {
	struct pcmcia_device *dev;
	struct dev_node_t node;
	unsigned char *mem;
	struct ib_net_modem_t *modem;
	struct timer_list timer;
	struct timer_list release;
	enum ib_pcmcia_ifstate ifstate;
	unsigned int sequence, feedback;
	int rx_echunk, rx_ibuf, rx_nbuf;
	int tx_echunk, tx_jbuf;
	int irx, jtx;
	int nreset;
};

/**
 * ib_pcmcia_dev_info - identifier for this driver.
 */
static dev_info_t ib_pcmcia_dev_info = "ibut_cs";

/**
 * ib_pcmcia_blit - the selected blit function.
 */
static void (*ib_pcmcia_blit)(unsigned char*, unsigned char*, int);

/**
 * ib_pcmcia_timer - poll hardware.
 * @_local: device private state.
 */
static void ib_pcmcia_timer(unsigned long _local)
{
	struct ib_pcmcia_local_t *local = (struct ib_pcmcia_local_t*) _local;

	ib_net_schedule(local->modem);
	local->timer.expires = jiffies + msecs_to_jiffies(interval);
	local->timer.function = ib_pcmcia_timer;
	local->timer.data = _local;
	add_timer(&local->timer);
}

/**
 * ib_pcmcia_blit16 - blit chunks (32 bytes) 16 bits at a time.
 * @_dst: destination buffer.
 * @_src: source buffer.
 * @nchunk: chunk count.
 */
static void ib_pcmcia_blit16(unsigned char *_dst, unsigned char *_src,
		int nchunk)
{
	uint16_t *dst = (uint16_t*) _dst;
	uint16_t *src = (uint16_t*) _src;

	while (nchunk--) {
		dst[0] = src[0];
		dst[1] = src[1];
		dst[2] = src[2];
		dst[3] = src[3];
		dst[4] = src[4];
		dst[5] = src[5];
		dst[6] = src[6];
		dst[7] = src[7];
		dst[8] = src[8];
		dst[9] = src[9];
		dst[10] = src[10];
		dst[11] = src[11];
		dst[12] = src[12];
		dst[13] = src[13];
		dst[14] = src[14];
		dst[15] = src[15];
		dst += 16;
		src += 16;
	}
}

/**
 * ib_pcmcia_blit32 - blit chunks (32 bytes) 32 bits at a time.
 * @_dst: destination buffer.
 * @_src: source buffer.
 * @nchunk: chunk count.
 */
static void ib_pcmcia_blit32(unsigned char *_dst, unsigned char *_src,
		int nchunk)
{
	uint32_t *dst = (uint32_t*) _dst;
	uint32_t *src = (uint32_t*) _src;

	while (1 < nchunk) {
		dst[0] = src[0];
		dst[1] = src[1];
		dst[2] = src[2];
		dst[3] = src[3];
		dst[4] = src[4];
		dst[5] = src[5];
		dst[6] = src[6];
		dst[7] = src[7];
		dst[8] = src[8];
		dst[9] = src[9];
		dst[10] = src[10];
		dst[11] = src[11];
		dst[12] = src[12];
		dst[13] = src[13];
		dst[14] = src[14];
		dst[15] = src[15];
		dst += 16;
		src += 16;
		nchunk -= 2;
	}
	if (nchunk) {
		dst[0] = src[0];
		dst[1] = src[1];
		dst[2] = src[2];
		dst[3] = src[3];
		dst[4] = src[4];
		dst[5] = src[5];
		dst[6] = src[6];
		dst[7] = src[7];
	}
}

/**
 * ib_pcmcia_release - release configuration.
 * @_local: device private state.
 */
static void ib_pcmcia_release(unsigned long _local)
{
	struct ib_pcmcia_local_t *local = (struct ib_pcmcia_local_t *) _local;
	struct pcmcia_device *link = local->dev;

	del_timer_sync(&local->timer);
	ib_net_flush(local->modem);
	iounmap(local->mem);
	pcmcia_disable_device(link);
}

/**
 * ib_pcmcia_detach - detach device.
 * @link: device state.
 */
static void ib_pcmcia_detach(struct pcmcia_device *link)
{
	struct ib_pcmcia_local_t *local = link->priv;

	del_timer_sync(&local->release);
	ib_pcmcia_release((unsigned long) local);
	if (local->modem)
		ib_net_deregister(local->modem);
	kfree(local);
}

/**
 * ib_pcmcia_reset - reset hardware protocol.
 *     called under ib_lock
 * @local: device private state.
 */
static void ib_pcmcia_reset(struct ib_pcmcia_local_t *local)
{
	struct ib_net_modem_t *modem = local->modem;
	unsigned char *mem = local->mem;
	unsigned int sequence, feedback;

	mem[OFFSET_PC_MAGIC2] = 0x00;
	mem[OFFSET_PC_MAGIC1] = 0x00;
	sequence = mem[OFFSET_UT_SEQUENCE];
	feedback = mem[OFFSET_UT_FEEDBACK];
	if (local->nreset)
		local->sequence = (feedback + 1) & 0xff;
	else
		local->sequence = (feedback + 7) & 0xff;
	local->feedback = sequence;
	mem[OFFSET_PC_SEQUENCE] = local->sequence;
	mem[OFFSET_PC_FEEDBACK] = local->feedback;
	mem[OFFSET_PC_IRX] = local->irx = 0;
	mem[OFFSET_PC_JTX] = local->jtx = 0;
	mem[OFFSET_PC_PACKET] = modem->stats.tx_packets & 0xff;
	mem[OFFSET_PC_STATUS] = modem->pc_status;
	mem[OFFSET_PC_MAGIC1] = 0xac;
	mem[OFFSET_PC_MAGIC2] = 0x02;
	local->ifstate = IFSTATE_WAIT;
	local->nreset += 1;
}

/**
 * ib_pcmcia_check - check hardware consistent.
 *     called under ib_lock
 * @local: device private state.
 */
static int ib_pcmcia_check(struct ib_pcmcia_local_t *local)
{
	unsigned char *mem = local->mem;

	if (mem[OFFSET_PC_MAGIC1] != 0xac || mem[OFFSET_PC_MAGIC2] != 0x02
			|| mem[OFFSET_PC_SEQUENCE] != local->sequence
			|| mem[OFFSET_PC_FEEDBACK] != local->feedback
			|| mem[OFFSET_PC_IRX] != local->irx
			|| mem[OFFSET_PC_JTX] != local->jtx) {
		DEBUG(8, "ib-pcmcia: check %02X %02X %02X %02X "
				"%02X %02X %02X %02X %02X %02X %02X %02X\n",
				mem[OFFSET_PC_MAGIC1], mem[OFFSET_PC_MAGIC2],
				mem[OFFSET_PC_PACKET], mem[OFFSET_PC_STATUS],
				mem[OFFSET_PC_SEQUENCE], local->sequence,
				mem[OFFSET_PC_FEEDBACK], local->feedback,
				mem[OFFSET_PC_IRX], local->irx,
				mem[OFFSET_PC_JTX], local->jtx);
		ib_pcmcia_reset(local);
		return 1;
	}
	return 0;
}

/**
 * ib_pcmcia_normal - normal operation; transfer chunks.
 *     called under ib_lock
 * @local: device private state.
 */
static void ib_pcmcia_normal(struct ib_pcmcia_local_t *local)
{
	struct ib_net_modem_t *modem = local->modem;
	struct ib_net_radio_t *radio = (struct ib_net_radio_t*) modem->rx_buf;
	unsigned char *mem = local->mem;
	int irx, jrx, rx_ichunk, rx_nchunk;
	int itx, jtx, tx_jchunk, tx_nchunk;
	int nchunk, nbuf;
	int packet;

	if (ib_pcmcia_check(local))
		return;
	if (mem[OFFSET_UT_MAGIC1] != 0xac || mem[OFFSET_UT_MAGIC2] != 0x02
			|| mem[OFFSET_UT_SEQUENCE] != local->feedback
			|| mem[OFFSET_UT_FEEDBACK] != local->sequence) {
		DEBUG(8, "ib-pcmcia: normal %02X %02X "
				"%02X %02X %02X %02X\n",
				mem[OFFSET_UT_MAGIC1], mem[OFFSET_UT_MAGIC2],
				mem[OFFSET_UT_SEQUENCE], local->feedback,
				mem[OFFSET_UT_FEEDBACK], local->sequence);
		mem[OFFSET_PC_IRX] = local->irx = 0;
		mem[OFFSET_PC_JTX] = local->jtx = 0;
		local->ifstate = IFSTATE_WAIT;
		return;
	}
	packet = modem->stats.tx_packets & 0xff;
	ib_net_ut_status(modem, mem[OFFSET_UT_STATUS]);
	if (mem[OFFSET_PC_PACKET] != packet)
		mem[OFFSET_PC_PACKET] = packet;
	if (mem[OFFSET_PC_STATUS] != modem->pc_status)
		mem[OFFSET_PC_STATUS] = modem->pc_status;
	rx_ichunk = mem[OFFSET_UT_ICHUNK];
	rx_nchunk = mem[OFFSET_UT_NCHUNK];
	tx_jchunk = mem[OFFSET_PC_JCHUNK];
	tx_nchunk = mem[OFFSET_PC_NCHUNK];
	if (rx_ichunk == 0 || rx_nchunk == 0)
		goto failed;
	if (tx_jchunk == 0 || tx_nchunk == 0)
		goto failed;
	if (tx_jchunk <= rx_ichunk && rx_ichunk < tx_jchunk + tx_nchunk)
		goto failed;
	if (rx_ichunk <= tx_jchunk && tx_jchunk < rx_ichunk + rx_nchunk)
		goto failed;
	if (0x80 < rx_ichunk + rx_nchunk || 0x80 < tx_jchunk + tx_nchunk)
		goto failed;
	while (1) {
		irx = local->irx;
		if (mem[OFFSET_PC_IRX] != irx)
			goto failed;
		jrx = mem[OFFSET_UT_JRX];
		if (2 * rx_nchunk <= jrx)
			goto failed;
		if (irx == jrx)
			break;
		if (jrx < irx)
			nchunk = jrx - irx + 2 * rx_nchunk;
		else
			nchunk = jrx - irx;
		if (rx_nchunk < nchunk)
			goto failed;
		if (rx_nchunk <= irx)
			irx -= rx_nchunk;
		if (rx_nchunk - irx < nchunk)
			nchunk = rx_nchunk - irx;
		if (local->rx_echunk == 0) {
			ib_pcmcia_blit(modem->rx_buf,
					mem + ((rx_ichunk + irx) << 5), 1);
			if (radio->word[0] & FLAG_EXTENSION)
				goto failed;
			nbuf = IB_NET_NBUF(radio->word[0], radio->word[1]);
			if (nbuf < IB_NET_RADIO_HEAD)
				goto failed;
			if (IB_NET_RADIO_HEAD + ETH_DATA_LEN < nbuf)
				goto failed;
			local->irx += 1;
			if (local->irx == 2 * rx_nchunk)
				local->irx = 0;
			mem[OFFSET_PC_IRX] = local->irx;
			if (nbuf <= 32) {
				ib_net_rx_parse(modem, nbuf);
				continue;
			}
			local->rx_echunk = ((nbuf + 31) >> 5) - 1;
			local->rx_ibuf = 32;
			local->rx_nbuf = nbuf;
			continue;
		}
		if (local->rx_echunk < nchunk)
			nchunk = local->rx_echunk;
		ib_pcmcia_blit(modem->rx_buf + local->rx_ibuf,
				mem + ((rx_ichunk + irx) << 5), nchunk);
		local->irx += nchunk;
		if (2 * rx_nchunk <= local->irx)
			local->irx -= 2 * rx_nchunk;
		mem[OFFSET_PC_IRX] = local->irx;
		local->rx_echunk -= nchunk;
		local->rx_ibuf += nchunk << 5;
		if (local->rx_echunk == 0)
			ib_net_rx_parse(modem, local->rx_nbuf);
	}
	while (1) {
		if (local->tx_echunk == 0) {
			nbuf = ib_net_tx_prepare(modem);
			if (nbuf == 0)
				break;
			modem->stats.tx_bytes += nbuf;
			modem->stats.tx_packets += 1;
			local->tx_echunk = (nbuf + 31) >> 5;
			local->tx_jbuf = 0;
		}
		itx = mem[OFFSET_UT_ITX];
		if (2 * tx_nchunk <= itx)
			goto failed;
		jtx = local->jtx;
		if (mem[OFFSET_PC_JTX] != jtx)
			goto failed;
		if (jtx < itx)
			nchunk = itx - jtx - tx_nchunk;
		else
			nchunk = itx - jtx + tx_nchunk;
		if (nchunk < 0)
			goto failed;
		if (nchunk == 0)
			break;
		if (tx_nchunk <= jtx)
			jtx -= tx_nchunk;
		if (tx_nchunk - jtx < nchunk)
			nchunk = tx_nchunk - jtx;
		if (local->tx_echunk < nchunk)
			nchunk = local->tx_echunk;
		ib_pcmcia_blit(mem + ((tx_jchunk + jtx) << 5),
				modem->tx_buf + local->tx_jbuf, nchunk);
		local->tx_echunk -= nchunk;
		local->tx_jbuf += nchunk << 5;
		local->jtx += nchunk;
		if (2 * tx_nchunk <= local->jtx)
			local->jtx -= 2 * tx_nchunk;
		mem[OFFSET_PC_JTX] = local->jtx;
	}
	return;
failed:
	DEBUG(1, "ib-pcmcia: normal failed\n");
	ib_pcmcia_reset(local);
}

/**
 * ib_pcmcia_wait - hardware reset wait; get hardware address.
 *     called under ib_lock
 * @local: device private state.
 */
static void ib_pcmcia_wait(struct ib_pcmcia_local_t *local)
{
	struct ib_net_modem_t *modem = local->modem;
	unsigned char *mem = local->mem;

	if (ib_pcmcia_check(local))
		return;
	local->feedback = mem[OFFSET_UT_SEQUENCE];
	mem[OFFSET_PC_FEEDBACK] = local->feedback;
	local->rx_echunk = 0;
	local->tx_echunk = 0;
	mem[OFFSET_PC_IRX] = local->irx = 0;
	mem[OFFSET_PC_JTX] = local->jtx = 0;
	if (mem[OFFSET_UT_MAGIC1] != 0xac || mem[OFFSET_UT_MAGIC2] != 0x02
			|| mem[OFFSET_UT_FEEDBACK] != local->sequence) {
		DEBUG(8, "ib-pcmcia: wait %02X %02X %02X %02X\n",
				mem[OFFSET_UT_MAGIC1], mem[OFFSET_UT_MAGIC2],
				mem[OFFSET_UT_FEEDBACK], local->sequence);
		return;
	}
	ib_net_addr(modem, mem + OFFSET_UT_ETHERNET);
	local->ifstate = IFSTATE_NORMAL;
	ib_pcmcia_normal(local);
}

/**
 * ib_pcmcia_poll - poll hardware according to protocol.
 *     called under ib_lock
 * @_local: device private state.
 */
static void ib_pcmcia_poll(void *_local)
{
	struct ib_pcmcia_local_t *local = _local;

	switch (local->ifstate) {
	case IFSTATE_RESET:
		ib_pcmcia_reset(local);
		break;
	case IFSTATE_WAIT:
		ib_pcmcia_wait(local);
		break;
	case IFSTATE_NORMAL:
		ib_pcmcia_normal(local);
		break;
	}
}

/**
 * ib_pcmcia_open - open device.
 *     called under ib_lock
 * @_local: device private state.
 */
static int ib_pcmcia_open(void *_local)
{
	struct ib_pcmcia_local_t *local = _local;
	struct ib_net_modem_t *modem = local->modem;
	unsigned char *mem = local->mem;

	DEBUG(9, "ib-pcmcia: +open\n");
	mem[OFFSET_PC_STATUS] = (modem->pc_status |= STATUS_READY);
	mem[OFFSET_PC_PACKET] = modem->stats.tx_packets & 0xff;
	DEBUG(9, "ib-pcmcia: -open\n");
	return 0;
}

/**
 * ib_pcmcia_rx_parse -
 *     called under ib_lock
 * @_local: device private state.
 */
static void ib_pcmcia_rx_parse(void *_local)
{
	struct ib_pcmcia_local_t *local = _local;
	struct ib_net_modem_t *modem = local->modem;
	unsigned char *mem = local->mem;

	ib_net_ut_status(modem, mem[OFFSET_UT_STATUS]);
	mem[OFFSET_PC_STATUS] = modem->pc_status;
	mem[OFFSET_PC_PACKET] = modem->stats.tx_packets & 0xff;
}

/**
 * ib_pcmcia_close - close device; release in a momemt.
 *     called under ib_lock
 * @_local: device private state.
 */
static int ib_pcmcia_close(void *_local)
{
	struct ib_pcmcia_local_t *local = _local;
	struct ib_net_modem_t *modem = local->modem;
	unsigned char *mem = local->mem;

	if (pcmcia_dev_present(local->dev)) {
		mem[OFFSET_PC_STATUS] = (modem->pc_status &= ~STATUS_READY);
		mem[OFFSET_PC_PACKET] = modem->stats.tx_packets & 0xff;
	}
	return 0;
}

/**
 * ib_pcmcia_net_driver - PCMCIA driver callback functions.
 */
static struct ib_net_driver_t ib_pcmcia_net_driver = {
	.poll = ib_pcmcia_poll,
	.tx_timeout = NULL,
	.open = ib_pcmcia_open,
	.rx_parse = ib_pcmcia_rx_parse,
	.close = ib_pcmcia_close,
};

/**
 * ib_pcmcia_config - PCMCIA config.
 * @link: device state.
 */
static int ib_pcmcia_config(struct pcmcia_device *link)
{
	struct ib_pcmcia_local_t *local = link->priv;
	struct net_device *netdev;
	struct ib_net_modem_t *modem;
	struct tuple_t tuple;
	union cisparse_t parse;
	unsigned char buf[64];
	int version;
	struct config_info_t conf;
	struct cistpl_cftable_entry_t *cfg;
	struct cistpl_cftable_entry_t dflt = { 0 };
	struct win_req_t req;
	int err;

	tuple.DesiredTuple = CISTPL_MANFID;
	tuple.Attributes = 0;
	tuple.TupleData = buf;
	tuple.TupleDataMax = sizeof(buf);
	tuple.TupleOffset = 0;
	err = pcmcia_get_first_tuple(link, &tuple);
	if (err)
		goto failed1;
	err = pcmcia_get_tuple_data(link, &tuple);
	if (err)
		goto failed1;
	err = pcmcia_parse_tuple(link, &tuple, &parse);
	if (err)
		goto failed1;
	version = parse.manfid.card;
	DEBUG(1, "ib-pcmcia: card version %d\n", version);
	tuple.DesiredTuple = CISTPL_CONFIG;
	tuple.Attributes = 0;
	tuple.TupleData = buf;
	tuple.TupleDataMax = sizeof(buf);
	tuple.TupleOffset = 0;
	err = pcmcia_get_first_tuple(link, &tuple);
	if (err)
		goto failed1;
	err = pcmcia_get_tuple_data(link, &tuple);
	if (err)
		goto failed1;
	err = pcmcia_parse_tuple(link, &tuple, &parse);
	if (err)
		goto failed1;
	link->conf.ConfigBase = parse.config.base;
	link->conf.Present = parse.config.rmask[0];
	err = pcmcia_get_configuration_info(link, &conf);
	if (err)
		goto failed1;
	tuple.DesiredTuple = CISTPL_CFTABLE_ENTRY;
	err = pcmcia_get_first_tuple(link, &tuple);
	if (err)
		goto failed1;
	while (1) {
		if (pcmcia_get_tuple_data(link, &tuple))
			goto next_entry;
		if (pcmcia_parse_tuple(link, &tuple, &parse))
			goto next_entry;
		cfg = &parse.cftable_entry;
		if (cfg->flags & CISTPL_CFTABLE_DEFAULT)
			dflt = *cfg;
		if (cfg->index == 0)
			goto next_entry;
		link->conf.ConfigIndex = cfg->index;
		if (cfg->flags & CISTPL_CFTABLE_AUDIO) {
			link->conf.Attributes |= CONF_ENABLE_SPKR;
			link->conf.Status = CCSR_AUDIO_ENA;
		}
		if (cfg->vcc.present & (1 << CISTPL_POWER_VNOM)) {
			if (conf.Vcc != cfg->vcc.param[CISTPL_POWER_VNOM]
					/ 10000)
				goto next_entry;
		} else if (dflt.vcc.present & (1 << CISTPL_POWER_VNOM)) {
			if (conf.Vcc != dflt.vcc.param[CISTPL_POWER_VNOM]
					/ 10000)
				goto next_entry;
		}
		if (cfg->vpp1.present & (1 << CISTPL_POWER_VNOM))
			link->conf.Vpp = cfg->vpp1.param[CISTPL_POWER_VNOM]
					/ 10000;
		else if (dflt.vpp1.present & (1 << CISTPL_POWER_VNOM))
			link->conf.Vpp = dflt.vpp1.param[CISTPL_POWER_VNOM]
					/ 10000;
		break;
next_entry:
		err = pcmcia_get_next_tuple(link, &tuple);
		if (err)
			goto failed1;
	}
	err = pcmcia_request_configuration(link, &link->conf);
	if (err) {
		DEBUG(1, "ib-pcmcia: pcmcia_request_configuration failed\n");
		goto failed1;
	}
	req.AccessSpeed = 0;
	req.Attributes = WIN_MEMORY_TYPE_CM | WIN_DATA_WIDTH_16 | WIN_ENABLE;
	req.Base = 0;
	req.Size = version < 2 ? 0x1000 : 0x2000;
	link->win = NULL;
	err = pcmcia_request_window(&link, &req, &link->win);
	if (err) {
		DEBUG(1, "ib-pcmcia: pcmcia_request_window failed\n");
		goto failed2;
	}
	local->mem = (unsigned char*) ioremap(req.Base, req.Size);
	if (local->mem == NULL) {
		DEBUG(1, "ib-pcmcia: ioremap failed\n");
		err = -EIO;
		goto failed3;
	}
	err = ib_net_register(&netdev);
	if (err)
		goto failed4;
	modem = netdev->priv;
	strcpy(local->node.dev_name, netdev->name);
	ib_net_addr(modem, local->mem + OFFSET_UT_ETHERNET);
	modem->driver = &ib_pcmcia_net_driver;
	modem->pdriver = local;
	local->modem = modem;
	local->ifstate = IFSTATE_RESET;
	local->nreset = 0;
	init_timer(&local->timer);
	local->timer.expires = jiffies + HZ / 4;
	local->timer.function = ib_pcmcia_timer;
	local->timer.data = (unsigned long) local;
	add_timer(&local->timer);
	local->node.major = local->node.minor = 0;
	local->node.next = NULL;
	return 0;
failed4:
	iounmap(local->mem);
failed3:
failed2:
	pcmcia_disable_device(link);
failed1:
	DEBUG(1, "ib-pcmcia: config failed\n");
	return err;
}

/**
 * ib_pcmcia_probe - probe device.
 * @dev: device state.
 */
static int ib_pcmcia_probe(struct pcmcia_device *link)
{
	struct ib_pcmcia_local_t *local;

	local = kmalloc(sizeof(struct ib_pcmcia_local_t), GFP_KERNEL);
	if (local == NULL)
		return -ENOMEM;
	local->dev = link;
	link->priv = local;
	link->conf.Attributes = 0;
	link->conf.IntType = INT_MEMORY_AND_IO;
	link->conf.Status = 0;
	link->conf.Vpp = 0;
	init_timer(&local->release);
	local->release.function = &ib_pcmcia_release;
	local->release.data = (unsigned long) local;
	local->modem = NULL;
	return ib_pcmcia_config(link);
}

static struct pcmcia_device_id ib_pcmcia_table[] = {
	PCMCIA_DEVICE_MANF_CARD(0x02e3, 0x0001),
	PCMCIA_DEVICE_MANF_CARD(0x02e3, 0x0002),
	PCMCIA_DEVICE_PROD_ID12("Kyocera Corporation", "Access Card",
			0xaba79a14, 0xd66618ce),
	PCMCIA_DEVICE_NULL,
};

MODULE_DEVICE_TABLE(pcmcia, ib_pcmcia_table);

/**
 * ib_pcmcia_driver - PCMCIA driver struct.
 */

static struct pcmcia_driver ib_pcmcia_driver = {
	.owner = THIS_MODULE,
	.drv = {
		.name = ib_pcmcia_dev_info,
	},
	.probe = ib_pcmcia_probe,
	.remove = ib_pcmcia_detach,
	.id_table = ib_pcmcia_table,
};

/**
 * ib_pcmcia_init - ib-pcmcia module init.
 */
static int __init ib_pcmcia_init(void)
{
	if (io16)
		ib_pcmcia_blit = ib_pcmcia_blit16;
	else
		ib_pcmcia_blit = ib_pcmcia_blit32;
	pcmcia_register_driver(&ib_pcmcia_driver);
	return 0;
}

/**
 * ib_pcmcia_exit - ib-pcmcia module exit.
 */
static void __exit ib_pcmcia_exit(void)
{
	pcmcia_unregister_driver(&ib_pcmcia_driver);
}

module_init(ib_pcmcia_init);
module_exit(ib_pcmcia_exit);

MODULE_DESCRIPTION("iBurst compatible PCMCIA driver");
MODULE_LICENSE("GPL");
